/* -*- mode:c -*- */
/** @file
 ** @brief OFI RDMA Transport implementation
 **/
/*
 * Copyright (C) 2020 Cray Computer GmbH
 * Copyright (C) 2021 HPE Switzerland GmbH
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include "maestro/logging.h"
#include "maestro/i_ofi.h"
#include "maestro/i_memlock.h"
#include "maestro/i_memory.h"
#include "maestro/i_pool_manager_protocol.h"
#include "maestro/i_pool_manager_registry.h"

#include "transport.h"
#include "transport_rdma.h"

#include "attributes/maestro-schema.h"

#include <rdma/fabric.h> 
#include <rdma/fi_domain.h>
#include <rdma/fi_rma.h>
#include <unistd.h>
#include <sys/mman.h>
#include <pthread.h>
#include <string.h>

/* simplify logging */
#define DEBUG(...) LOG_DEBUG(MSTRO_LOG_MODULE_TRANSP,__VA_ARGS__)
#define INFO(...)  LOG_INFO(MSTRO_LOG_MODULE_TRANSP,__VA_ARGS__)
#define WARN(...)  LOG_WARN(MSTRO_LOG_MODULE_TRANSP,__VA_ARGS__)
#define ERR(...)   LOG_ERR(MSTRO_LOG_MODULE_TRANSP,__VA_ARGS__)

/** hashing of memory registrations is done on an aggregate key: CDO ID, OFI domain and maybe OFI endpoint.
 * Be sure to make * padding 0, and set EP to NULL if FI_MR_ENDPOINT is not set in the domain
 * attributes */
struct mstro_transport_mreg_table_key {
  struct mstro_cdo_id id;      /**< CDO ID for which this registration is done */
  struct fid_domain *domain;   /**< domain in which this registration is valid */
  struct fid_ep     *ep;       /**< endpoint (if FI_MR_ENDPOINT is set) */
};

/*  https://ofiwg.github.io/libfabric/v1.10.1/man/fi_mr.3.html */
struct mstro_transport_mreg_table_entry {
  UT_hash_handle hh;           /**< hash on CDO id */
  struct mstro_transport_mreg_table_key key; /**< hash key */

  uint64_t refcount; /**< number of tickets issued for this ID */
  void *addr; /**< the raw-ptr address registered */
  uint64_t len; /**< length of the memory region */
  struct fid_mr *mr_reg; /**< the OFI memory registration for this CDO's raw-ptr */
};

struct mstro_transport_rdma_pending_entry {
  UT_hash_handle hh;      /**< hash on CDO id */
  struct mstro_cdo_id id; /**< key */

  pthread_mutex_t lock;
  pthread_cond_t cond;
  int done;
};

/** table of memory registrations for transport purposes. Protected by @ref g_mstro_transport_mreg_table_lock */
static struct mstro_transport_mreg_table_entry *g_mstro_transport_mreg_table = NULL;
/** Lock protecting @ref g_mstro_transport_mreg_table */
static pthread_mutex_t g_mstro_transport_mreg_table_lock = PTHREAD_MUTEX_INITIALIZER;
/** condition var to signal change on @ref g_mstro_transport_mreg_table */
static pthread_cond_t g_mstro_transport_mreg_table_cond = PTHREAD_COND_INITIALIZER;

static struct mstro_transport_rdma_pending_entry *g_mstro_transport_rdma_pending_table = NULL;
static pthread_mutex_t g_mstro_transport_rdma_pending_table_mtx = PTHREAD_MUTEX_INITIALIZER;

/** find the registry entry for endpoint,cdoid, assuming lock on table is already held by caller */
static mstro_status
mstro_transport_rdma__mreg_entry_find__locked(struct mstro_endpoint *ep,
                                              const struct mstro_cdo_id *cdoid,
                                              struct mstro_transport_mreg_table_entry **regentry_p)
{
  assert(ep!=NULL);
  
  struct mstro_transport_mreg_table_key key;
  memset(&key,0,sizeof(struct mstro_transport_mreg_table_key));
  
  key.id = *cdoid;
  key.domain = ep->domain;
  key.ep = ep->fi->domain_attr->mr_mode & FI_MR_ENDPOINT ? ep->ep : NULL;

  HASH_FIND(hh, g_mstro_transport_mreg_table,
            &key, sizeof(struct mstro_transport_mreg_table_key),
	    *regentry_p);
  if (!*regentry_p) {
    return MSTRO_NOENT;
  } else {
    return MSTRO_OK;
  }
}

/** find the registry entry for (APPID,CDOID) */
static mstro_status
mstro_transport_rdma__mreg_entry_find(mstro_app_id appid,
                                      const struct mstro_cdo_id *cdoid,
                                      struct mstro_transport_mreg_table_entry **regentry_p)
{
  assert(appid!=MSTRO_APP_ID_INVALID);
  assert(cdoid!=NULL);
  if(regentry_p==NULL)
    return MSTRO_INVOUT;
  
  struct mstro_pm_app_registry_entry *e;
  mstro_status status = mstro_pm_app_lookup(appid, &e);
  if(e==NULL) {
    ERR("App %" PRIappid " not in app table\n", appid);
    return MSTRO_FAIL;
  } 

  int err = pthread_mutex_lock(&g_mstro_transport_mreg_table_lock);
  if (err) {
    ERR("Couldn't lock mutex on mreg table\n");
	return MSTRO_FAIL;
  }
  status = mstro_transport_rdma__mreg_entry_find__locked(e->ep, cdoid, regentry_p);
  
  err = pthread_mutex_unlock(&g_mstro_transport_mreg_table_lock);
  if (err) {
    ERR("Couldn't unlock mutex on mreg table\n");
    status=MSTRO_FAIL;
  }
  return status;
}


mstro_status  
mstro_transport_rdma_src_execute(mstro_cdo src, Mstro__Pool__TransferTicket* ticket)
{
  if (src == NULL || ticket == NULL)
    return MSTRO_INVARG;

/* XXX we will be ignoring Mamba for now */

  if ((uintptr_t)src->raw_ptr % (size_t) sysconf(_SC_PAGESIZE) != 0) {
    WARN ("CDO data is not pagesize aligned, fall-back on other transport\n");
    return MSTRO_INVARG;
  }

  mstro_status status = MSTRO_OK;
  struct mstro_transport_mreg_table_entry* regentry = NULL;
  struct mstro_cdo_id* cdo_id_p = &(src->gid);

  struct mstro_pm_app_registry_entry *e;
  status = mstro_pm_app_lookup(ticket->ofi->dstid->id, &e);
  if(e==NULL) {
    ERR("Target %zu not in app table\n", ticket->ofi->dstid);
    return MSTRO_FAIL;
  } 

  status = mstro_transport_rdma__mreg_entry_find__locked(e->ep, cdo_id_p, &regentry);

  if (regentry) {
    INFO("another listener on existing mreg\n");
    regentry->refcount++;
    goto FILL_TICKET_FROM_REGISTRATION;
  }
  
FRESH_REGISTRATION:
  /* make a fresh entry */
  regentry = calloc(1, sizeof(struct mstro_transport_mreg_table_entry)); /* to make sure we have 0-initialized key padding */
  if(regentry==NULL) {
    ERR("Cannot allocate new mreg entry\n");
    status=MSTRO_NOMEM;
    goto BAILOUT_UNLOCK;
  }

  memset(&regentry->key, 0, sizeof(struct mstro_transport_mreg_table_key));
  
  regentry->key.id = *cdo_id_p;
  regentry->key.domain = e->ep->domain;
  regentry->key.ep = e->ep->fi->domain_attr->mr_mode & FI_MR_ENDPOINT ? e->ep->ep : NULL;

  regentry->refcount=1;

  {
    struct mstro_transport_datalen dl;
    status = mstro_transport__src_datalen_get(src, ticket, &dl);
    if (status != MSTRO_OK) {
      ERR("Couldn't get CDO data and size for RDMA transport (status: %s)\n", mstro_status_description(status));
      goto BAILOUT_UNLOCK;
    }
    regentry->addr=dl.data;
    regentry->len=dl.len;

    if (mstro_memlock(dl.data, dl.len) != MSTRO_OK) {
      ERR("Couldn't lock raw-ptr for RDMA transport of CDO (%s)\n", src->name);
      status=MSTRO_FAIL;
      goto BAILOUT_UNLOCK;
    }

    /*  register this address for OFI */
    struct fid_mr *mr;
    uint64_t requested_key = e->ep->fi->domain_attr->mr_mode & FI_MR_PROV_KEY ? 0 : mstro_memory_new_key();
    int err = fi_mr_reg(e->ep->domain, dl.data, dl.len,
                        FI_REMOTE_READ, 0, requested_key, 0, &mr, NULL);
    if(err) {
      ERR("Couldn't register memory region for RDMA transport (err: %d)\n", err);
      status=MSTRO_FAIL;
      goto BAILOUT_UNLOCK;
    }
    regentry->mr_reg = mr;
    WITH_CDO_ID_STR(
	idstr, &(regentry->key.id),
	DEBUG("Adding CDO %s (id %s) to mreg table\n",
	      src->name, idstr););

    HASH_ADD(hh, g_mstro_transport_mreg_table,
	     key, sizeof(struct mstro_transport_mreg_table_key),
	     regentry);
  }

FILL_TICKET_FROM_REGISTRATION:
  {
    uint8_t* mr_key=NULL;
    size_t keysize=0;
    uint64_t mr_addr=(uint64_t)regentry->addr;
    status = mstro_mr_key_get(e->ep->fi, regentry->mr_reg, &mr_key, &keysize, &mr_addr);
    if(status!=MSTRO_OK) {
      ERR("Failed to obtain MR key: %d (%s)\n", 
	  status, mstro_status_description(status));
      goto BAILOUT_UNLOCK;
    }

    ticket->ofi->h->mr_key.len = keysize;
    ticket->ofi->h->mr_key.data = mr_key;
    ticket->ofi->h->single = mr_addr;
    //DEBUG("Ticket: key-len %zu, data %p, addr %p\n", ticket->ofi->h->mr_key.len, ticket->ofi->h->mr_key.len, ticket->ofi->h->single );
  }

BAILOUT_UNLOCK:
  {
    int err = pthread_mutex_unlock(&g_mstro_transport_mreg_table_lock);
    if (err) {
      ERR("Couldn't unlock mutex on mreg table\n");
      status=MSTRO_FAIL;
    }
  }

  return status;
}

/* "Bottom half" execution handler: called by mstro_pc__handle_transfer_completed */
mstro_status
mstro_transport_rdma_src_execute_bh(Mstro__Pool__TransferCompleted *tc)
{
  if (tc == NULL) {
    ERR ("Transfer completed msg is null\n");
    return MSTRO_INVARG;
  }
  mstro_status status = MSTRO_OK;

  struct mstro_transport_mreg_table_entry* regentry = NULL;
  struct mstro_cdo_id cdoid = { .qw[0] = tc->srccdoid->qw0,
                                .qw[1] = tc->srccdoid->qw1,
                                .local_id = tc->srccdoid->local_id };

  struct mstro_pm_app_registry_entry *e;
  status = mstro_pm_app_lookup(tc->dstid->id, &e);
  if(e==NULL) {
    ERR("Target %" PRIappid " not in app table\n", tc->dstid->id);
    return MSTRO_FAIL;
  } 

  status = mstro_transport_rdma__mreg_entry_find__locked(e->ep, &cdoid, &regentry);

  if (!regentry) {
    DEBUG("RDMA transfer completion message does not require memory registration cleanup on source, likely 0-length\n");
    status=MSTRO_NOENT;
    goto BAILOUT_UNLOCK;
  }

  regentry->refcount--;
  DEBUG ("A client is done using memory region (start: %p) used for RDMA\n", regentry->addr);

  if (regentry->refcount > 0) 
    goto BAILOUT_UNLOCK; 

  DEBUG ("Unregistering memory region (start: %p) used for RDMA\n", regentry->addr);

  int err = fi_close((struct fid*)regentry->mr_reg);
  if (err) {
    ERR("Couldn't unregister memory region for RDMA transport (err: %d)\n", err);
    status=MSTRO_FAIL;
    goto BAILOUT_UNLOCK;
  }

  status = mstro_memunlock(regentry->addr, regentry->len);
  if (status != MSTRO_OK) {
    ERR("Couldn't unlock raw-ptr for RDMA transport\n");
    goto BAILOUT_UNLOCK;
  }

  HASH_DEL(g_mstro_transport_mreg_table, regentry);
  free(regentry);
  if (HASH_COUNT(g_mstro_transport_mreg_table) == 0) {
    err = pthread_cond_signal(&g_mstro_transport_mreg_table_cond);
    if (err) {
      ERR("Couldn't signal mreg table is empty\n");
      status = MSTRO_FAIL;
    }
  }

BAILOUT_UNLOCK:
  err = pthread_mutex_unlock(&g_mstro_transport_mreg_table_lock);
  if (err) {
    ERR("Couldn't unlock mutex on mreg table\n");
    status=MSTRO_FAIL;
  }
  
  return status;
}
  
mstro_status
mstro_transport_rdma_dst_execute(mstro_cdo cdo_dst, Mstro__Pool__TransferTicket* ticket)
{
  if (cdo_dst == NULL || ticket == NULL) {
    ERR ("cdo or ticket is null\n");
    return MSTRO_INVARG;
  }

/* XXX we will be ignoring Mamba for now */

  mstro_status status = MSTRO_OK;
  uint64_t len = ticket->data_size;
  assert(len > 0); /* this should have been checked before */
  int fresh_alloc_route=false;
  struct mstro_transport_rdma_cb_args* closure=NULL;

  if (cdo_dst->raw_ptr != NULL) {
    if ((uintptr_t)cdo_dst->raw_ptr % (size_t) sysconf(_SC_PAGESIZE) != 0) {
      ERR ("Can't memlock an unaligned raw_ptr\n");
      return MSTRO_INVARG;
    }
    if (mstro_memlock(cdo_dst->raw_ptr, len) != MSTRO_OK) {
      ERR("Couldn't lock raw-ptr for RDMA transport (read) of CDO (%s)\n", cdo_dst->name);
      return MSTRO_FAIL;
    }
  } else { 
    cdo_dst->raw_ptr = mstro_alloc_lockable(len, true); /* locking immediately */
    DEBUG("Doing fresh_alloc\n");
    fresh_alloc_route=1;
    assert (cdo_dst->raw_ptr!=NULL);
  }

  /*  register this address for OFI */
  struct mstro_pm_app_registry_entry *app_entry = NULL;
  status = mstro_pm_app_lookup(ticket->srcid->id, &app_entry);
  if(app_entry==NULL) {
    ERR("Target %zu not in app table\n", ticket->srcid);
	return MSTRO_FAIL;
  } else {
    DEBUG("Target found in app table\n");
  }   

  /* DEBUG("Doing ticket inspection\n"); */
  /* DEBUG("ticket: %p, ofi %p, h %p, keylen %zu, data %p\n", */
  /*       	  ticket, ticket->ofi, ticket->ofi->h, ticket->ofi->h->mr_key.len, ticket->ofi->h->mr_key.data); */

  size_t keysize = ticket->ofi->h->mr_key.len;
  uint8_t* tmp = (uint8_t*)(ticket->ofi->h->mr_key.data);
  uint64_t mr_addr = ticket->ofi->h->single;
  uint64_t mr_key=*(uint64_t*)ticket->ofi->h->mr_key.data;
  struct fid_mr *mr=NULL;

  DEBUG("Doing memory registration\n");

/*  switch(keysize) {
    case 1: { uint8_t  x; memcpy(&x,tmp, sizeof(x)); mr_key = x; break; }
    case 2: { uint16_t x; memcpy(&x,tmp, sizeof(x)); mr_key = x; break; }
    case 4: { uint32_t x; memcpy(&x,tmp, sizeof(x)); mr_key = x; break; }
    case 8: {             memcpy(&mr_key,tmp, sizeof(mr_key));   break; }
    default: {
      ERR("Unhandled MR key length -- likely needs OFI based translation\n");
      return MSTRO_UNIMPL;
    }
  }
*/

  uint64_t requested_key = app_entry->ep->fi->domain_attr->mr_mode & FI_MR_PROV_KEY ? 0 : mstro_memory_new_key();
  int err = fi_mr_reg(app_entry->ep->domain, cdo_dst->raw_ptr, len,
                  FI_READ, 0, requested_key, 0, &mr, NULL);
  if (err) {
	ERR("Couldn't register memory region for RDMA transport (err: %d, %s)\n", 
	    err, fi_strerror(-err));
	return MSTRO_FAIL;
  }

  void* local_buf_mr_desc = fi_mr_desc(mr);
  mstro_ofi_msg_context ctx=NULL;
  status = mstro_ofi__msg_context_create(&ctx, NULL, true);
  if(status!=MSTRO_OK) {
    ERR("Failed to init msg context for RDMA op: %d (%s)\n",
        status, mstro_status_description(status));
    return status;
  } 

  DEBUG("Doing closure creation\n");
  closure = malloc(sizeof(struct mstro_transport_rdma_cb_args));
  if(closure==NULL) {
    ERR("Failed to alloc RDMA transport closure\n");
    return MSTRO_NOMEM;
  }
  closure->mr = mr;
  closure->fresh_alloc_route = fresh_alloc_route;
  closure->raw_ptr = cdo_dst->raw_ptr;
  closure->len = len;
  closure->name = strdup(cdo_dst->name);
  if( closure->name == NULL) {
    ERR("Failed to allocate RDMA transport closure\n");
    free(closure);
    return MSTRO_NOMEM;
  }
  closure->srccdoid.qw[0]    = ticket->srccdoid->qw0;
  closure->srccdoid.qw[1]    = ticket->srccdoid->qw1;
  closure->srccdoid.local_id = ticket->srccdoid->local_id;
  closure->dstcdoid = cdo_dst->gid;
  closure->appid = ticket->srcid->id;

  DEBUG("Doing event creation\n");
  status = mstro_event_create(g_transport_rdma_edom, mstro_transport_rdma_cb,
                              (void*)closure, mstro_transport_rdma_dtor, 0, &(ctx->ev));
  if (status != MSTRO_OK) {
    ERR("Couldn't create event (%s)\n", mstro_status_description(status));
    return MSTRO_FAIL;
  } else {
    mstro_event_id eid;
    status = mstro_event_id_get(ctx->ev, &eid);
    if (status != MSTRO_OK) {
      DEBUG("Failed to query event id (event @%p)\n", ctx->ev);
      return MSTRO_FAIL;
    }
    DEBUG("Created event id %" PRIx64 " to monitor RDMA read completion.\n", eid);
  }

  /* make a fresh entry (for completion) */
  struct mstro_transport_rdma_pending_entry* regentry = malloc(sizeof(struct mstro_transport_rdma_pending_entry));
  if(regentry==NULL) {
    ERR("Cannot allocate new transport completion entry\n");
    return MSTRO_NOMEM;
  }
  memcpy(&(regentry->id),
         &(cdo_dst->gid),
	 sizeof(struct mstro_cdo_id));
  pthread_mutex_init(&(regentry->lock), NULL);
  pthread_cond_init(&(regentry->cond), NULL);
  regentry->done = 0;

  int ret=pthread_mutex_lock(&g_mstro_transport_rdma_pending_table_mtx);
  if(ret!=0) {
    ERR("Failed to lock pending RDMA op table\n");
    status=MSTRO_FAIL;
    goto BAILOUT;
  } 
  // entry on which we'll wait on for completion
  HASH_ADD(hh, g_mstro_transport_rdma_pending_table,
           id, sizeof(struct mstro_cdo_id), regentry);
  ret=pthread_mutex_unlock(&g_mstro_transport_rdma_pending_table_mtx);
  if(ret!=0) {
    ERR("Failed to unlock pending RDMA op table\n");
    status=MSTRO_FAIL;
    goto BAILOUT;
  }

  int num_retries = 3;
 RETRY_RDMA_TRANSPORT_READ: ;
  ret = fi_read(app_entry->ep->ep, cdo_dst->raw_ptr, len, local_buf_mr_desc,
                    app_entry->addr, mr_addr, mr_key, ctx);

  if(ret==-FI_EAGAIN) {
    INFO("RDMA read for CDO transport (%s) needs to be retried\n", cdo_dst->name);
    sleep(1);
    if(num_retries-->0)
      goto RETRY_RDMA_TRANSPORT_READ;
  }
  if(ret<0) {
    ERR("Failed to do RDMA read for CDO transport (%s) (err: %s)\n", cdo_dst->name,fi_strerror(-ret));
	status = MSTRO_FAIL;
  }
BAILOUT:
  return status;
}

/** callback triggered when RDMA completes on receiving side of CDO transfer */ 
void
mstro_transport_rdma_cb(mstro_event ev, void* closure)
{
  mstro_event foo=ev; // avoid unused warning
  struct mstro_transport_rdma_cb_args* args = closure;
  WITH_CDO_ID_STR(
      srcidstr, &(args->srccdoid),
      WITH_CDO_ID_STR(
          dstidstr, &(args->dstcdoid),
          DEBUG("Callback is finishing completion msg of CDO `%s`-> `%s`\n",
                srcidstr, dstidstr);););

  mstro_status status;
  args->status = MSTRO_OK;

  // unregister
  int err = fi_close((struct fid*)args->mr);
  if (err) {
    ERR("Couldn't unregister memory region for RDMA transport (err: %d)\n", err);
	args->status = MSTRO_FAIL;
  }

  // send completion to PM and sender
  WITH_CDO_ID_STR(
      idstr, &(args->srccdoid),
      INFO("Sending transfer completion msg of CDO `%s` to PM and sender\n",
           idstr););
  status = mstro_pc__transport_send_completion(args->appid,
                                               &args->srccdoid,
                                               &args->dstcdoid,
                                               1); // RDMA src always wants completion
  if (status != MSTRO_OK) {
    ERR("Couldn't send completion for RDMA transport of CDO `%s`\n", args->name);
    args->status = status;
  }

  // unlock
  DEBUG("fresh_alloc_route = %d\n", args->fresh_alloc_route);
  if (!args->fresh_alloc_route) {
    status = mstro_memunlock(args->raw_ptr, args->len);
    if (status != MSTRO_OK) {
      ERR("Couldn't unlock raw-ptr for RDMA transport of CDO `%s`\n", args->name);
	  args->status = status;
    }
  } 
  else
    mstro_unlock_lockable (args->raw_ptr, args->len);

  struct mstro_transport_rdma_pending_entry* regentry;

  int ret=pthread_mutex_lock(&g_mstro_transport_rdma_pending_table_mtx);
  if(ret!=0) {
    ERR("Failed to lock pending RDMA op table\n");
    args->status=MSTRO_FAIL;
    return;
  } 
  HASH_FIND(hh, g_mstro_transport_rdma_pending_table,
            &(args->dstcdoid), sizeof(struct mstro_cdo_id), regentry);

  ret=pthread_mutex_unlock(&g_mstro_transport_rdma_pending_table_mtx);
  if(ret!=0) {
    ERR("Failed to unlock pending RDMA op table\n");
    args->status=MSTRO_FAIL;
    return;
  }
  if (!regentry) {
    ERR("RDMA transfer completion message does not map to a documented transfer on destination\n");
    args->status=MSTRO_FAIL;
    return;
  }

  // mark readiness for (the transport thread and then) the local pool
  // there is only one waiter (the transport thread) for one CDO transfer completion
/*  regentry->done = 1;
  err = pthread_cond_signal(&(regentry->cond));  
  if (err) {
    ERR("Couldn't signal (errno: %d)\n", err);
    args->status = MSTRO_FAIL;
  }
*/
  /* indicate that CDO is filled */
  mstro_cdo cdo;
  mstro_cdo__find_cdo(&(args->dstcdoid), &cdo);
  args->status = mstro_cdo__mark_transfer_complete(cdo);
  if(args->status!=MSTRO_OK) {
    ERR("Failed to indicate completion of incoming CDO: %d (%s)\n",
        args->status, mstro_status_description(args->status));
    return;
  }
  args->status=MSTRO_OK;
  
  DEBUG("RDMA transport for CDO `%s` successful\n", args->name);
}

/** cleanup function for RDMA callback arguments of completion events */
void
mstro_transport_rdma_dtor(void* closure)
{
  struct mstro_transport_rdma_cb_args* args = closure;

  struct mstro_pm_app_registry_entry *e;
  args->status = mstro_pm_app_lookup(args->appid, &e);
  if(args->status!=MSTRO_OK) {
    ERR("Sender app %zu not in app table\n", args->appid);
    return;
  } 

  struct mstro_transport_rdma_pending_entry* regentry;
  int ret=pthread_mutex_lock(&g_mstro_transport_rdma_pending_table_mtx);
  if(ret!=0) {
    ERR("Failed to lock pending RDMA op table\n");
    args->status=MSTRO_FAIL;
    return;
  } 
  HASH_FIND(hh, g_mstro_transport_rdma_pending_table,
            &(args->dstcdoid), sizeof(struct mstro_cdo_id), regentry);
  ret=pthread_mutex_unlock(&g_mstro_transport_rdma_pending_table_mtx);
  if(ret!=0) {
    ERR("Failed to unlock pending RDMA op table\n");
    args->status=MSTRO_FAIL;
    return;
  }
  // clean up completion hash entry
  if (!regentry) {
    ERR("RDMA transfer completion message does not map to a documented transfer on destination\n");
    args->status = MSTRO_FAIL;
  }
  pthread_mutex_destroy(&(regentry->lock));
  pthread_cond_destroy(&(regentry->cond));
  HASH_DEL(g_mstro_transport_rdma_pending_table, regentry);
  free(regentry);
  
  
  // clean up closure structure elements
  //  free(args);
  free((void*)args->name); // strdup'ed
  
}



mstro_status
mstro_transport_rdma_finalize(void)
{
  mstro_status s;
  mstro_cleanup_waitup(s,
                       g_mstro_transport_mreg_table,
                       &g_mstro_transport_mreg_table_lock,
                       &g_mstro_transport_mreg_table_cond,
                       "RDMA transport memory registration");
  if(s!=MSTRO_OK) {
    struct mstro_transport_mreg_table_entry *entry, *tmp;
    HASH_ITER(hh, g_mstro_transport_mreg_table, entry, tmp) {
      WITH_CDO_ID_STR(idstr, &entry->key.id, {
		      WARN("Still have mreg entry %p: CDO %s\n", entry, idstr);
                      });
    }
  }
  return s;
}
